#
# This is a part of software for the algorithm described in
#
# Kuzborskij, I., Carlucci, F. M., Caputo, B.
# "When Naive Bayes Nearest Neighbours Meet Convolutional Neural Networks".
# In: Computer Vision and Pattern Recognition (CVPR), 2016 IEEE Conference on.
#
# Contact the author: ilja.kuzborskij [at] idiap.ch
#                     http://idiap.ch/~ikuzbor
#
# ----------------------------------------------------------------------------------
#
# Copyright (c) 2015 Idiap Research Institute, http://www.idiap.ch/
# Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
#
# This file is part of STOML3.
#
# STOML3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# STOML3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with STOML3. If not, see <http://www.gnu.org/licenses/>.
#

import sys
import numpy as np
from random import shuffle
import random
from scipy.sparse import csr_matrix, issparse
import sklearn
from sklearn.base import BaseEstimator, ClassifierMixin
from smooth_ml3_ctypes import POINTER, c_float, c_uint32, c_bool, \
    get_cu_train_time_ms, get_cu_test_time_ms, get_cu_aer_track, \
    train_cu_smooth_ml3, test_cu_smooth_ml3, destroy_cu_smooth_ml3, cu_load_model, get_model as cu_get_model

from common import *

class MinibatchDenseOneShotLoader(object):
   """ In-memory training helper. """
   def __init__(self, passes, _X, _y, _tags, transpose=False, minibatch_size=2500, shuffle=True):
       self.passes = passes
       self._X = _X
       self._y = _y-1
       self._tags = _tags
       self.ix = np.arange(len(self._y))

       self.transpose = transpose
       self.minibatch_size = minibatch_size

       self.cursor = 0
       self.do_shuffle = shuffle

       if self.do_shuffle:
          self.shuffle()

   def loader(self):
       if self.passes > 0:
           start = self.cursor
           end = min(self.cursor + self.minibatch_size, self._X.shape[0])

           mat = self._X[self.ix[start:end]]
           labels = self._y[self.ix[start:end]]

           if self._tags is not None and len(self._tags) > 0:
              tags = self._tags[self.ix[start:end]]
           else:
              tags = None

           if self.transpose:
               mat = mat.T

           if end == self._X.shape[0]:
               self.passes -= 1
               self.cursor = 0
               if self.do_shuffle:
                  self.shuffle()
           else:
               self.cursor += self.minibatch_size

           return mat, labels, tags
       else:
           return None, None, None

   def shuffle(self):
       shuffle(self.ix)

class SmoothML3(BaseEstimator, ClassifierMixin):
   """ Implements STOML3 (Stochastic Multiclass Locally-Linear Learning) algorithm."""

   def __init__(self, passes=10, lambda_=0.001, n=10, t0=0.0, p=2.0, L=1.0, minibatch_size=100,
                 model_path=None, save_model_every_t=0, model_tag=None, nbnl_mode=False, ml3_hand=None):
      """ Constructor.

      passes -- number of passes to perform on the training set (default 10)
      lambda_ -- regularization parameter (default 0.001)
      n -- number of locally-linear models per class (default 10)
      t0 -- learning rate tuning parameter, the bigger it is, the smaller is the starting learning rate (default 0)
      L -- Lipschitz constant guess for objective function (default 1)
      minibatch_size -- minibatch size when training from in-memory dataset (default 100)
      model_path -- where to store intermediate models
      save_model_every_t -- save model every t iterations (default 0 -- do not save)
      model_tag -- specific ID that is used when saving model files
      nbnl_mode -- run in NBNL (Naive Bayes Non-linear Learning) mode (default False)
      ml3_hand -- reserved
      """
      self.passes = passes
      self.lambda_ = lambda_
      self.gauss_init_std = 1.0
      self.t0 = t0
      self.n = n
      self.p = p
      self.L = L
      self.minibatch_size = minibatch_size
      self.record_aer = True
      self.ml3_hand = ml3_hand
      self.model_path=model_path
      self.save_model_every_t=save_model_every_t
      self.model_tag = model_tag
      self.nbnl_mode = nbnl_mode

   def get_params(self, deep=True):
      return {'passes': self.passes, 'lambda_': self.lambda_, 't0': self.t0, 'n': self.n, 'p': self.p, 'L': self.L, 'minibatch_size': self.minibatch_size,
              'ml3_hand': self.ml3_hand, 'record_aer': self.record_aer,
              'model_path': self.model_path, 'model_tag': self.model_tag, 'save_model_every_t': self.save_model_every_t, 'nbnl_mode': self.nbnl_mode}

   def set_params(self, **parameters):
      for parameter, value in parameters.items():
         setattr(self, parameter, value)
      return self

   def fit(self, X, y):
      """ Trains STOML3 classifier.
      
      X -- the training set (num. examples x dimension)
      y -- labels in {1, 2, ...}
      """
      X = add_reg_bias(X)

      log = get_logger()

      loader = MinibatchDenseOneShotLoader(self.passes, X, y, None, transpose=True, minibatch_size=self.minibatch_size)
      train = train_cu_smooth_ml3

      destroy_cu_smooth_ml3(self.ml3_hand)
      self.ml3_hand = train(loader,
                            classes=len(np.unique(y)),
                            n=self.n,
                            L=self.L,
                            t0=self.t0,
                            p=self.p,
                            lambda_=self.lambda_,
                            gauss_init_std=self.gauss_init_std,
                            record_aer=self.record_aer,
                            model_path=self.model_path,
                            save_model_every_t=self.save_model_every_t,
                            model_tag=self.model_tag,
                            validation_loader=None,
                            nbnl=self.nbnl_mode)
      self.train_time_ms = get_cu_train_time_ms(self.ml3_hand)

   def fit_from_loader(self, loader, num_classes, validation_loader=None):
      """ Trains STOML3 classifier in out-of-core mode by retrieving mini-batches from a given loader.
      
      loader -- a loader to receive training data from*
      num_classes -- total number of classes expected
      validation_loader -- a loader providing a validation set

      *loader is an instance of a class of a following template:
      class ExampleLoader:
          def loader(self):
                ''' Provides a minibatch to the learner. Called every time after minibatch is consumed, unless
                returns None, None, None -- this marks the end of training.
     
                The minibatch consists of three elements: (X, y, nbnl_tags), where
                X -- examples (minibatch size x dimension)
                y -- labels in {1, 2, ...} (minibatch size)
                nbnl_tags -- a vector with elements in {0,1}, where 1 indicates the beginning of an example in NBNL sense (minibatch size).
                             Can be None if training is non-nbnl mode.
     
                Note also that the number of passes must also be controlled by this loader.
                '''
                pass

      See data.py, which contains CIFAR10 minibatch loader example (non-nbnl).
      """
      destroy_cu_smooth_ml3(self.ml3_hand)

      self.ml3_hand = train_cu_smooth_ml3(loader,
                                          classes=num_classes,
                                          n=self.n,
                                          L=self.L ,
                                          t0=self.t0,
                                          p=self.p,
                                          lambda_=self.lambda_,
                                          gauss_init_std=self.gauss_init_std,
                                          record_aer=self.record_aer,
                                          model_path=self.model_path,
                                          save_model_every_t=self.save_model_every_t,
                                          model_tag=self.model_tag,
                                          validation_loader=validation_loader,
                                          nbnl=self.nbnl_mode)
      self.train_time_ms = get_cu_train_time_ms(self.ml3_hand)

   def score_from_loader(self, loader, on_gpu=True, nbnl=False):
      """ Evaluates STOML3 classifier in out-of-core mode by retrieving mini-batches from a given loader.
      
      loader -- a loader to receive testing data from (see documentation of fit_from_loader for detail)
      on_gpu -- whether to test on GPU
      nbnl -- whether to test in NBNL mode
      """
      
      score = test_cu_smooth_ml3(self.ml3_hand, loader, on_gpu, nbnl)
      self.test_time_ms = get_cu_test_time_ms(self.ml3_hand)
      destroy_cu_smooth_ml3(self.ml3_hand)
      return score

   def score(self, X, y, on_gpu=True):
      """ Tests STOML3 classifier.
      
      X -- the testing set (num. examples x dimension)
      y -- labels in {1, 2, ...}
      """
      X = add_reg_bias(X)

      loader = MinibatchDenseOneShotLoader(1, X, y, None, transpose=True, minibatch_size=self.minibatch_size)
      test = test_cu_smooth_ml3

      score = test(self.ml3_hand, loader, on_gpu)
      self.test_time_ms = get_cu_test_time_ms(self.ml3_hand)

      #destroy_cu_smooth_ml3(self.ml3_hand)

      return score

   def get_model(self, k, dim):
      """ Retrieves model weights for a class k.
      
      k -- class index
      dim -- number of dimensions to retrieve
      """
      return cu_get_model(self.ml3_hand, k, self.n, dim)      

   def finish(self):
      """ Frees up resources used by STOML3. """
      destroy_cu_smooth_ml3(self.ml3_hand)

   def get_aer_track(self):
      """ Gets average error rate sequence. """
      return get_cu_aer_track(self.ml3_hand)
