#
# This is a part of software for the algorithm described in
#
# Kuzborskij, I., Carlucci, F. M., Caputo, B.
# "When Naive Bayes Nearest Neighbours Meet Convolutional Neural Networks".
# In: Computer Vision and Pattern Recognition (CVPR), 2016 IEEE Conference on.
#
# Contact the author: ilja.kuzborskij [at] idiap.ch
#                     http://idiap.ch/~ikuzbor
#
# ----------------------------------------------------------------------------------
#
# Copyright (c) 2015 Idiap Research Institute, http://www.idiap.ch/
# Written by Ilja Kuzborskij <ilja.kuzborskij@idiap.ch>
#
# This file is part of STOML3.
#
# STOML3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# STOML3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with STOML3. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from scipy.io import loadmat
from sklearn.preprocessing import StandardScaler
from common import *

class Cifar10MinibatchLoader(object):
    '''
    This class implements reading of batch files and on-the-fly splitting into minibatches.
    The method loader() is supposed to be called repeatedly by minibatch consumer until None, None is returned.
    init_scaler() must be called before first call to loader(), that is to initialize normalization parameters.
    '''

    def __init__(self, files, passes, minibatch_size=1024, scaler=None):
        self.filenames = sorted(files)
        self.n_files = len(self.filenames)
        self.i = 0

        self.passes = passes
        self.minibatch_size = minibatch_size

        self.X = None
        self.y = None
        self.current_file_size = 0
        self.cursor = 0
        self.scaler = scaler

    def init_scaler(self):
        mat = loadmat(self.filenames[0])
        X = mat['data'].astype(np.float32)
        self.scaler = StandardScaler(copy=False)
        self.scaler.fit_transform(X)

    def loader(self):
        if self.cursor >= self.current_file_size: # Depleted current file, will read a new one.
            if self.i == self.passes*self.n_files: # Or, if no files left to feed just pass None, None.
                return None, None, None

            if 0 == (self.i % self.n_files):
                get_logger().info('Pass: %s', (self.i / self.n_files)+1)

            mat = loadmat(self.filenames[self.i % self.n_files])

            self.X = mat['data']
            self.y = mat['labels'].ravel().astype(dtype=np.uint32)

            self.current_file_size = self.X.shape[0]
            self.cursor = 0
            self.i += 1


        # Getting next minibatch from current file
        start = self.cursor
        end = min(self.cursor+self.minibatch_size, self.current_file_size)
        X = self.X[start:end, :]
        y = self.y[start:end]
        self.cursor += self.minibatch_size

        # Normalizing the data
        X = X.astype(np.float32)
        self.scaler.transform(X)

        return X.T, y, None
